/*
 * Copyright (C) 2002
 * 	Hidetoshi Shimokawa. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *
 *	This product includes software developed by Hidetoshi Shimokawa.
 *
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id
 */

#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <dev/firewire/firewire.h>

#include <netinet/in.h>
#include <fcntl.h>
#include <stdio.h>
#include <err.h>
#include <stdlib.h>
#include <strings.h>

void
usage(void)
{
	printf("bus_mgm [-g gap_count] [-b pri_req] [-r]\n");
	printf("\t-g: broadcast gap_count by phy_config packet\n");
	printf("\t-b: set PRIORITY_BUDGET register on all supported nodes\n");
	printf("\t-r: bus reset\n");
	exit(0);
}

void
get_num_of_dev(int fd, struct fw_devlstreq *data)
{
	data->n = 64;
	if( ioctl(fd, FW_GDEVLST, data) < 0) {
       			err(1, "ioctl");
	}
	printf("%d devices\n", data->n);
}

u_int32_t
read_write_quad(int fd, struct fw_eui64 eui, u_int32_t addr_lo, int read, u_int32_t data)
{
        struct fw_asyreq *asyreq;
	u_int32_t *qld, res;

        asyreq = (struct fw_asyreq *)malloc(sizeof(struct fw_asyreq_t) + 16);
	asyreq->req.len = 16;
	asyreq->req.type = FWASREQEUI;
	asyreq->req.dst.eui = eui;
#if 0
	asyreq->pkt.mode.rreqq.dst = htons(0xffc0 | node);
#endif
	asyreq->pkt.mode.rreqq.tlrt = 0;
	if (read)
		asyreq->pkt.mode.rreqq.tcode = FWTCODE_RREQQ;
	else
		asyreq->pkt.mode.rreqq.tcode = FWTCODE_WREQQ;

	asyreq->pkt.mode.rreqq.dest_hi = htons(0xffff);
	asyreq->pkt.mode.rreqq.dest_lo = htonl(addr_lo);

	qld = (u_int32_t *)&asyreq->pkt;
	if (!read)
		asyreq->pkt.mode.wreqq.data = data;

	if (ioctl(fd, FW_ASYREQ, asyreq) < 0) {
       		err(1, "ioctl");
	}
	res = qld[3];
	free(asyreq);
	if (read)
		return ntohl(res);
	else
		return 0;
}
void
send_phy_config(int fd, int root_node, int gap_count)
{
        struct fw_asyreq *asyreq;

	asyreq = (struct fw_asyreq *)malloc(sizeof(struct fw_asyreq_t) + 12);
	asyreq->req.len = 12;
	asyreq->req.type = FWASREQNODE;
	asyreq->pkt.mode.ld[0] = 0;
	asyreq->pkt.mode.ld[1] = 0;
	asyreq->pkt.mode.common.tcode = FWTCODE_PHY;
	if (root_node >= 0)
		asyreq->pkt.mode.ld[1] |= htonl((root_node & 0x3f) << 24 | 1 << 23);
	if (gap_count >= 0)
		asyreq->pkt.mode.ld[1] |= htonl(1 << 22 | (gap_count & 0x3f) << 16);
	asyreq->pkt.mode.ld[2] = ~asyreq->pkt.mode.ld[1];

	printf("send phy_config root_node=%d gap_count=%d\n",
						root_node, gap_count);

	if (ioctl(fd, FW_ASYREQ, asyreq) < 0) {
       		err(1, "ioctl");
	}
}

void
set_pri_req(int fd, int pri_req)
{
	struct fw_devlstreq data;
	u_int32_t max, reg, old;
	int i;

	get_num_of_dev(fd, &data);
#define BUGET_REG 0xf0000218
	for (i = 0; i < data.n; i++) {
		if (!data.status[i])
			continue;
		reg = read_write_quad(fd, data.eui[i], BUGET_REG, 1, 0);
		printf("%d %08x:%08x, %08x",
			data.dst[i], data.eui[i].hi, data.eui[i].lo, reg);
		if (reg > 0 && pri_req >= 0) {
			old = (reg & 0x3f);
			max = (reg & 0x3f00) >> 8;
			if (pri_req > max)
				pri_req =  max;
			printf(" 0x%x -> 0x%x\n", old, pri_req);
			read_write_quad(fd, data.eui[i], BUGET_REG, 0, pri_req);
		} else {
			printf("\n");
		}
	}
}

int
main(argc, argv)
        int argc;
        char *argv[];
{
	char devname[] = "/dev/fw1";
	int fd, tmp;

	if ((fd = open(devname, O_RDWR)) < 0)
		err(1, "open");

	if (argc < 2) {
		usage();
	}

	argv++;
	argc--;
	while (argc > 0) {
		if (strcmp(*argv, "-g") == 0) {
			/* gap count */
			argv++;
			argc--;
			if (argc > 0) {
				tmp = strtoul(*argv, (char **)NULL, 0);
				argv++;
				argc--;
				send_phy_config(fd, -1, tmp);
			} else {
				usage();
			}
		} else if (strcmp(*argv, "-b") == 0) {
			argv++;
			argc--;
			tmp = -1;
			if (argc > 0) {
				tmp = strtoul(*argv, (char **)NULL, 0);
				argv++;
				argc--;
			} else {
				usage();
			}
			set_pri_req(fd, tmp);
		} else if (strcmp(*argv, "-r") == 0) {
			argv++;
			argc--;
			/* bus reset */
			if(ioctl(fd, FW_IBUSRST, &tmp) < 0) {
                       		err(1, "ioctl");
			}
		}
	}
	return 0;
}
